/*
	leakmap.h
	<c> Risto Holopainen 2021
	
	unstable logistic map
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef LEAKMAP_H
#define LEAKMAP_H

#include <cmath>
#include <deque>

using std::deque;


namespace leakage
{

const double r_const = 4.001;

inline double map(double x)
{
	return r_const*x*(1-x);
}


struct counter
{
	unsigned int n = 0;
	double x = 0;
	deque <double> dq;
};


counter n_iterations(double x0)
{
	counter cc;
	double u = x0;
	unsigned int k = 0;
	while(u > 0 && u < 1)
		{
		u = map(u);
		cc.dq.push_back(u);
		++k;
		}
	if(k < 50)	// expand symmetrically
		for(unsigned int j=1; j<k; j += 2)
			cc.dq.push_front(cc.dq[j]);
	
	cc.n = k;								// number of elements id deque
	double w = (u - 1) / (.25*r_const - 1);	// normalized to [0,1]
	cc.x = w;
	return cc;
}


deque<double> expandeq(const deque<double> &q)
{
	// interpolate between entries
	const int N = q.size();
	int m;
	double ds, w, x;
	const double bias = 0.5;
	deque <double> p;
	p.push_back(0);

	const int fade_in = 25;
	for(int i=0; i<fade_in; i++)
		{
		w = (double) i/fade_in;
		x = w * (q[0] - bias);
		p.push_back(x);
		}

	for(int i=0; i<N-1; i++)
		{
		ds = fabs(q[i+1] - q[i]);
		if(ds > 0.04 && ds < 0.08)
			m = 3;
		else if (ds > 0.08 && ds < 0.16)
			m = 5;
		else if(ds > 0.16 && ds < 0.32)
			m = 8;
		else if(ds > 0.32 && ds < 0.64)
			m = 13;
		else if(ds > .64)
			m = 21;
		else
			m = 2;
		
		for(int j=0; j<m; j++)
			{
			w = (double) j/m;
			x = (1-w) * q[i] + w * q[i+1] - bias;
			p.push_back(x);	
			}
		}

	const int fade_out = 36;
	for(int i=0; i<fade_out; i++)
		{
		w = (double) i/fade_out;
		x = (1-w) * (q[N-1] - bias);
		p.push_back(x);
		}

	return p;
}


}	// end namespace


#endif // def. leakmap.h
