/*
	batir.cpp
	Copyright Risto Holopainen 2021

	first stage generator for the composition
	bâtir et démolir
	(the material then needs to be demolished by running demolir.cpp)


	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef M_PI
#define M_PI 3.14159265358979323846 // instead of #define _USE_MATH_DEFINES
#endif

#include <cstdio>
#include <cmath>
#include <sndfile.h>
#include "quadmap.h"

/*
	generate raw house material to be later demolished and reconstructed
	by a binopsie prog
*/


double sinusoid(double omega)
{
	static double phi=0;
	double x = sin(phi);
	phi += omega;
	return x;
}


double stoch(double p)
{
	static double x=1/13.;
	double r = fabs(p)+3;
	x = r*x*(1-x);
	return x;
}

double constance(double p)
{
	return p;
}


typedef double (*fp) (double);

typedef struct
{
	double L;
	double R;
}
stereo;


enum op {Sine, Stoch, Konst};
const char *opcodes[3] = {"sines", "stochastic", "constant"};


class synt
{
	public:
		synt();
		stereo gen(op how);
		void param(double p);

	private:
		double p;
		double W, omega;
		double phi;
		double x, y, u, v;
		stereo z;
		double P1, P2;
		qmap QM;
		quadosc QOS;
};


synt::synt()
{
	W = 2*M_PI/48000.;
	omega = W;
	phi = 0;
	x = 0.1;
	u = 0.5;
	v = 0.1;
	P1 = P2 = 0.5;
	QOS.frq(440*W);
	QM.param(0.5, 0.5);
	QM.init(0.5, 0.5);
}


void synt::param(double b)
{
	p = b;
	omega = W*b;
	P2 = P1;
	P1 = (fabs(p) < 1) ? fabs(p) : 0.5*P2;

	QM.param(P1, P2);
	QM.init(0.2, 0.5);	// init-value
	QOS.frq(omega);
}

stereo synt::gen(op how)
{
	switch (how)
		{
		case Sine :
			{
			QOS.gen(x, y);
			break;
			}
		case Stoch :
			{
			QM.iterate(u, v);
			x = norm_range(u);
			y = norm_range(v);
			break;
			}
		case Konst :
			{
			x = p;
			y = -p;
			break;
			}
		default :
			x = y = 0;
		}
	z.L = x;
	z.R = y;
	return z;
}



struct edifice
{
	double dur;
	op art;
	double val;
};




class build
{
	public:
		build();
		edifice brick(int) const;
		int size() const;

	private:
		enum {N=90, T=45};
		double s[N];
		edifice ed[N];
};


build::build()
{
	for(int i=0; i<N; i++)
		{
		s[i] = (double) T / (i+1);
		}

	// permutate
	double frq;
	op artist;


	for(int i=0, k=1, m; i<N; i++)
		{
		ed[i].dur = s[k];
		m = (k/2)%5;
		artist = (m < 3) ? Sine : Konst;
		if(i%4 == 3)
			artist = Stoch;
		ed[i].art = artist;

		if(artist == Sine)
			{
			frq = 24*(k+1) + .25*k*k;
			ed[i].val = frq;
			}
		else
			ed[i].val = (k%15)/7. - 1;

		k += 11;
		if(k >= N)
			k -= N;
		}

}

int build::size() const
{
	return (int) this->N;
}


edifice build::brick(int n) const
{
	// return duration, material [n]
	if(n >= N || n < 0)
		printf("bad brick index (%d)!\n", n);

	return ed[n];

}



void materialize()
{
	build cottage;
	edifice X;
	const int N = cottage.size();
	unsigned int T=0, k=0;
	int n=0;
	const int sr = 48000;
	const int L = 1 << 8;
	double buf[2*L];

	SNDFILE *ouf;
	SF_INFO head;
	head.samplerate = sr;
	head.channels = 2;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_16);
	ouf = sf_open("batir.wav", SFM_WRITE, &head);
	synt SNYT;
	stereo sig;

	while(n < N)
		{
		X = cottage.brick(n);
		T = sr * X.dur;
		k = 0;
		SNYT.param(X.val);

		while(k < T)
			{
			for(int i=0; i<2*L; i += 2, k++)
				{
				sig = SNYT.gen(X.art);
				buf[i  ] = .995*sig.L;
				buf[i+1] = .995*sig.R;
				}

			sf_write_double(ouf, buf, 2*L);
			}
		++n;
		}

	sf_close(ouf);
}




int main()
{
	build cottage;
	edifice tt;
	int n=90, i=0, p;
	int qq[3] = {0,0,0};

	while(i < n)
		{
		tt = cottage.brick(i);
		p = (int) tt.art;
		++qq[p];
		printf("%2d \t %6.2f \t op %s \t (%5.3f)\n", i, tt.dur, opcodes[p], tt.val);
		++i;
		}

	printf("\n---------\n");
	for(int i=0; i<3; i++)
		printf("%2d %s\n", qq[i], opcodes[i]);


	materialize();
	return 0;
}




