/*
	fm23
	Copyright Risto Holopainen 2021

	dual, triple FM, and "simple" cross-coupled FM
	some sparse documentation available at
	https://ristoid.net/modular/nonlin_osc.html

	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "../util/biquad2.h"


struct stereo
{
	double L;
	double R;
};


struct params
{
	double carr_1;
	double carr_2;
	double omega_1;
	double omega_2;
	double J;
	double K;
	double M;
	double A;
};



class dual_fm
{
	public:
		dual_fm(int sr);
		stereo gen();
		void set(const params &p);
		params get_params() const ;
		void set_J(double jj) {J=jj;}
		void set_K(double kk) {K=kk;}
		void set_M(double mm) {M=mm;}
		void set_C(double c_1, double c_2) {C1 = w*c_1; C2 = w*c_2;}
		void set_W(double w_1, double w_2) {W1 = w_1; W2 = w_2;}

	private:
		int sr;
		stereo x;
		double A;
		double phi1, phi2, theta1, theta2;
		double C1, C2, W1, W2, J, K, M;
		double w;
		dc_block DC1, DC2;
		double u, v, g1, g2, h1, h2;
};


dual_fm::dual_fm(int fs)
{
	sr = fs;
	phi1 = phi2 = theta2 = 0.0;
	theta1 = M_PI/6;
	w = 2*M_PI/sr;

	A = 0.93;
	C1 = w*110;
	C2 = w*110;
	W1 = w*110;
	W2 = w*110;
	J = 1.0;
	K = 1.0;
	M = 1.0;
}


void dual_fm::set(const params &p)
{
	C1 = w*p.carr_1;
	C2 = w*p.carr_2;
	W1 = w*p.omega_1;
	W2 = w*p.omega_2;
	J = p.J;
	K = p.K;
	M = p.M;
	A = p.A;
}

params dual_fm::get_params() const
{
	params p;
	p.J = J;
	p.K = K;
	p.M = M;
	p.A = A;
	p.carr_1 = C1 / w;
	p.carr_2 = C2 / w;
	p.omega_1 = W1 / w;
	p.omega_2 = W2 / w;
	return p;
}

stereo dual_fm::gen()
{
	u = A*sin(phi1);
	v = A*sin(phi2);

	g1 = sin(theta1);
	g2 = sin(theta2);
	h1 = sin(theta1 - phi1);
	h2 = sin(theta2 - phi2);

	phi1 += C1 * (1 + J*g1) + K*h1;
	phi2 += C2 * (1 + J*g2) + K*h2;

	theta1 += W1 * (1 + M*g2) - K*h1;
	theta2 += W2 * (1 + M*g1) - K*h2;

	x.L = DC1.filter(u);
	x.R = DC2.filter(v);

	return x;
}



class triple_fm
{
	public:
		triple_fm(int sr);
		stereo gen();
		void set_J(double jj) {J=jj;}
		void set_K(double kk) {K=kk;}
		void set_M(double mm) {M=mm;}
		void set_cf(const double cc[3])
			{ cf[0]=W*cc[0]; cf[1]=W*cc[1]; cf[2]=W*cc[2]; }
		void set_mf(const double mm[3])
			{mf[0]=W*mm[0]; mf[1]=W*mm[1]; mf[2]=W*mm[2]; }
		void get_params(double &J, double &K, double &M) const;
		void get_carriers(double cf[3]) const;
		void get_modulators(double mf[3]) const;

	private:
		int sr;
		double phi[3];
		double theta[3];
		double cf[3];		// carrier-frq
		double mf[3];		// modulator-frq
		double J, K, M;
		double A;
		double W;
		dc_block dc1, dc2;
		stereo x;
		double u, v;
		double a[3], b[3];
		double a1[3], b1[3];
		double a0, b0;
};


triple_fm::triple_fm(int fs)
{
	sr = fs;
	W = 2.*M_PI/sr;
	for(int i=0; i<3; i++)
		{
		phi[i] = 0;
		theta[i] = 0;
		cf[i] = W * 220;
		mf[i] = W * 220;
		a1[i] = b1[i] = 0;
		}

	A = 0.25;
	J = 1.0;
	K = 1.0;
	M = 1.0;
	u = v = 0;
}


stereo triple_fm::gen()
{
	for(int i=0; i<3; i++)	// lopass a,b
		{
		a0 = sin(theta[i]);
		b0 = sin(phi[i] - theta[i]);
		a[i] = .5 * (a0 + a1[i]);
		b[i] = .5 * (b0 + b1[i]);
		a1[i] = a0;
		b1[i] = b0;
		}

	for(int i=0, j=1; i<3; i++)
		{
		phi[i] += cf[i] * (1 + J*a[i]) - K*b[i];
		theta[i] += mf[i] * (1 + M*a[j]) + K*b[i];

		if(++j >= 3)
			j = 0;
		}

	u = 2 * sin(phi[0]) + sin(phi[1]);
	v = 2 * sin(phi[2]) + cos(phi[1]);
	x.L = A*dc1.filter(u);
	x.R = A*dc2.filter(v);

	return x;
}




void triple_fm::get_params(double &jj, double &kk, double &mm) const
{
	jj = J;
	kk = K;
	mm = M;
}




void triple_fm::get_carriers(double carr[3]) const
{

        //for(int i=0; i<3; carr[i] = cf[i]/W)
        //i++;
		for(int i=0; i<3;i++)carr[i] = cf[i]/W;

}


void triple_fm::get_modulators(double mods[3]) const
{
        //for(int i=0; i<3; mods[i] = mf[i]/W)
        //i++;
        for(int i=0; i<3; i++)mods[i] = mf[i]/W;

}



class simple_fm
{
	public:
		simple_fm(int sr, double cf=440, double mf=220);
		~simple_fm();
		void set(double carr, double mod, double ind, double fbk);
		void index(double I) {modindex=I;}
		double gen(double add_phase=0);

	private:
		int sr;
		double theta_c, theta_m;
		double w, omega_c, omega_m;
		double modindex;
		double x, y, fbx, fby;
		double fbind;
		one_pole <double> *lp;
		dc_block dc;
};

simple_fm::simple_fm(int fs, double cf, double mf)
{
	sr = fs;
	w = 2*M_PI/sr;
	modindex = 1.0;
	fbind = 0.2;
	omega_c = w*cf;
	omega_m = w*mf;
	x = y = fbx = fby = 0;
	lp = new one_pole<double> (.25, .75);
}

simple_fm::~simple_fm()
{
	lp = 0;
	delete lp;
}

void simple_fm::set(double carr, double mod, double ind, double fb)
{
	omega_c = w*carr;
	omega_m = w*mod;
	modindex = ind;
	fbind = fb;
}


double simple_fm::gen(double phase)
{
	theta_c += omega_c * (1 + fbind*fby);	// fbind > 1 --> oscil-death
	theta_m += omega_m * (1 + fbind*fbx);
	fbx = x;
	fby = y;
	x = sin(theta_m + lp->filter(phase));
	y = sin(theta_c + modindex * x);

	fbx += x;
	fby += y;

	return dc.filter(y);
}
