
/*
	follia.cpp
	Copyright Risto Holopainen 2021

	--folie a cinq--
	Generator of the composition La Follia

	 This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "folie.h"
#include "../util/azcr.h"
#include "../util/waveforms.h"
#include "../util/oscil.h"
#include "../util/delay_t.h"

#include <vector>
#include <sndfile.h>

#define VERBOSE
#undef VERBOSE

using std::vector;

typedef double fivect[5];

class five_osc
{
	public:
		five_osc(int sr);
		~five_osc();
		void run(const fivect frq, fivect x);


	private:
		int sz;
		int sr;
		double f1, f2, ds;	// freq regions
		double *wt_lo, *wt_hi;
		vector<oscil> osc_lo, osc_hi;

};

five_osc::five_osc(int fs)
{
	sr = fs;
	sz = 4097;
	wt_lo = new double [sz];
	wt_hi = new double [sz];
	f1 = 200;
	f2 = 900;
	ds = f2 - f1;

	waveform::make_band(wt_lo, sz, 3, 31);
	waveform::make_band(wt_hi, sz, 1, 3);

	for(int i=0; i<5; i++)
		{
		osc_lo.push_back(oscil(sr, wt_lo, sz));
		osc_hi.push_back(oscil(sr, wt_hi, sz));
		}
}

five_osc::~five_osc()
{
	wt_lo = 0;
	wt_hi = 0;
	delete wt_lo;
	delete wt_hi;
}


void five_osc::run(const double frq[5], fivect x)
{

	double a, b;

	for(int i=0; i<5; i++)
		{
		if(frq[i] < f1)
			{
			a = 1;
			b = 0;
			}
		else if(frq[i] >= f1 && frq[i] < f2)
			{
			a = (frq[i] - f1) / ds;
			b = 1 - a;
			}
		else
			{
			a = 0;
			b = 1;
			}

		x[i] = a * osc_lo[i].gen(frq[i]) + b * osc_hi[i].gen(frq[i]);
		}
}



void update_frq(fivect frq, const fivect fb, const fivect mod, const fivect amp)
{
	double maxfr = 0, ampsum=0;
	for(int i=0; i<5; i++)
		{
		ampsum += amp[i];
		if(fabs(frq[i]) > maxfr)
			maxfr = fabs(frq[i]);
		}

	const double F0 = 65 - ampsum;
	double g = 20 + 100. / (1 + .25*maxfr);
	double k = 5 + 2*sqrt(maxfr);

	for(int i=0; i<5; i++)
		{
		frq[i] = F0 * (1 + g*fb[i]) + k*mod[i] - 2*amp[i];
		if( fabs(frq[i]) > 9000 )
			frq[i] *= 0.5;
		}
}


void map(const loco &lfo, fivect mods)
{
		mods[0] = lfo.u;
		mods[1] = lfo.v;
		mods[2] = lfo.q;
		mods[3] = .5*(lfo.q - lfo.u);
		mods[4] = .5*(lfo.q - lfo.v);
}

double get_sd(vector<azcr> zv, fivect sd)
{
	double sum = 0;
	for (int i=0; i<5; i++)
		{
		sd[i] = zv[i].sd();
		sum += sd[i];
		}
	return sum;
}


class envelope
{
	public:
		envelope();
		double follow(double x);

	private:
		double y, phi, u, v;
		double h, b, c;
};

envelope::envelope()
{
	y = 0;
	phi = 0;
	h = 1/250.;
	b = 1 - 3E-04;
	c = (1-b)/4;
}

double envelope::follow(double x)
{
	u = fabs(x);
	phi += h * u;
	v = sin(phi);
	y = b * y + c * (3*u + v*v);

	return sqrt(y);
}


int main()
{
	const int sr = 48000;
	vector<azcr> ZCR;
	vector<delay<double> > yore;
	vector<envelope> anvil;
	const int N = 5;

	for(int i=0; i<N; i++)
		{
		ZCR.push_back(azcr(60-3*i));
		yore.push_back( delay<double>(750 + 133*i, 1/99.) );
		anvil.push_back(envelope());
		}

	five_osc hi_five(sr);
	//loco LFO{0.0, 0.0, 0.0, 0};
	loco LFO{};// ? MSV debug
	const unsigned char MAXFUN = 247;
	il_est_fou accum(MAXFUN);

	fivect x={0}, zc={0}, zd={0}, fb={0}, mods={0}, Frqs={0};
	double sum_z=0;

	SNDFILE *oup;
	SF_INFO head;
	head.samplerate = sr;
	head.channels = 5;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);

	oup = sf_open("la_follia.wav", SFM_WRITE, &head);

	const int L = 1 << 5;
	double buf[N*L], maxampL = 0.0, maxampR = 0.0;
	double norm = 1/2.;
	double maxfrq = 0;
	double beta;
	const int TT = 165*sr/L;
	const int mm[N] = {1,2,3,4,0}, nn[N] = {4,0,1,3,2};
	double lambda=0.05;
	const double h = 0.05;
	fivect sd = {0}, amp = {0};
	double sdsum=0, gg=1-2E-05, gain=1;
	const double espilon = 0.00245;

	for(int i=0; i<TT; i++)
		{
		for(int n=0, m=0; n<L; n++, m += head.channels)
			{
			sum_z = 0;
			map(LFO, mods);
			update_frq(Frqs, fb, mods, amp);
			hi_five.run(Frqs, x);
			beta = .9 + .75*LFO.v ;	// sensitive parameters

			if(LFO.count % 19 == 13)
				{
				sdsum = get_sd(ZCR, sd);
				lambda = (1 + 4.0/sdsum);
				#ifdef VERBOSE
					printf("sdsum: %.3f, lambda = %.3f, beta %.2f\n",
						sdsum, lambda, beta);
				#endif
				}

			for(int j=0; j<N; j++)
				{
				amp[j] = anvil[j].follow(x[j]);
				zc[j] = ZCR[j].rate(x[j]);
				zd[j] = yore[j].z(zc[mm[j]]);
				fb[j] += h * (lambda*zc[ nn[j] ]*(1 + zc[j]) - sd[j]*zd[j] - beta*fb[j]);
				sum_z += zc[j];
				if(fabs(Frqs[j]) > maxfrq)
					maxfrq = fabs(Frqs[j]);
				}

			for(int j=0; j<N; j++)
				{
				x[j] *= gain * amp[ nn[j] ];
				buf[m+j] = norm*x[j];
				}

			LFO = accum.run(espilon*sum_z);

			if(LFO.count >= MAXFUN)
				gain *= gg;

			if(fabs(buf[m]) > maxampL)
				maxampL = fabs(buf[n]);
			if(fabs(buf[m+1]) > maxampR)
				maxampR = fabs(buf[m+1]);
			}

		sf_write_double(oup, buf, N*L);
		if(maxfrq > sr/2)
			{
			printf("maxfrq insane: %.1f Hz\n", maxfrq);
			break;
			}
		}


	sf_close(oup);
	printf("maxamp: L %.2f, R %.2f\n", maxampL, maxampR);
	printf("count: %d\n\n", LFO.count);

	return 0;
}


