/*
	Quasi-periodic
	<c> Risto Holopainen 2021

	Generator for the composition The Quasi-periodic Table

	The output is a four channel audio file that you may not be able to
	open with just any audio player, but try importing it to audacity
	or some other editor.

	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#ifndef M_PI
#define M_PI 3.14159265358979323846 // instead of #define _USE_MATH_DEFINES
#endif


#include "constants.h"
#include "../util/oscil.h"
#include "../util/waveforms.h"
#include <sndfile.h>
#include <complex>

using std::complex;

typedef complex <double> Complex;


class slope
{
	public:
		slope();
		bool rising(double x);
		int profile(double x);	// -1, 0, +1

	private:
		double x1;
		double u1, u2;
		double eps;
};


slope::slope()
{
	eps = 0.003;
	u1 = u2 = 0;
	x1 = 0;
}

bool slope::rising(double x)
{
	double d = x - x1;
	x1 = x;
	return (d > 0);
}

int slope::profile(double x)
{
	double y = x - 2*u1 + u2;
	u2 = u1;
	u1 = x;

	int p;
	if (fabs(y) < eps)
		p = 0;

	else
		p = (y>0) ? 1 : -1;

	return p;
}


class quasiperiod
{
	public:
		quasiperiod();
		~quasiperiod() { printf("Quasiperiod: tuned %d times.\n", count) ; } ;
		void rotate(Complex cc[2]);
		bool tune(double w[6]);

	private:
		Complex w1, w2, w3, w4, w5, w6;
		double th1, th2, th3, th4;
		double h;
		double t;
		double g0, g1, g2;
		Complex I;
		int count;
		Complex Z_1, Z_2;
};


quasiperiod::quasiperiod()
{
	th1 = th2 = th3 = th4 = 0.0;
	h = 0.01;
	t = 0.0;
	I = Complex(0,1);
	w1 = I*sqrt(2);
	w2 = I*sqrt(11);
	w3 = I*sqrt(3);
	w4 = I*sqrt(13);
	w5 = I*sqrt(7);
	w6 = I*sqrt(5);
	count = 0;
	g1 = 0.5;
	g2 = 1/3.;
	g0 = 1.0;
}


void quasiperiod::rotate(Complex z[2])
{
	Complex u1 = g0*exp(w1*t);
	Complex u2 = g1*exp(w2*t) - g2*exp(w5*t);
	Complex v1 = g1*exp(w3*t) - g2*exp(w6*t);
	Complex v2 = g0*exp(w4*t);

	t += h;
	Complex p, q;

	p = u1 * z[0] + u2 * z[1];
	q = v1 * z[0] + v2 * z[1];

	z[0] += h*p;
	z[1] += h*q;
	Z_1 = z[0];
	Z_2 = z[1];
}


bool quasiperiod::tune(double ff[6])
{
	if(abs(Z_1) < 3/8. && abs(Z_2) < 3/8.)
		{
		w1 = I*ff[0] * 3.0;
		w2 = I*ff[1] * 3.0;
		w3 = I*ff[2] * 3.0;
		w4 = I*ff[3] * 3.0;
		w5 = I*ff[4] * 2.0;
		w6 = I*ff[5] * 2.0;
		++count;
		return true;
		}
	else
		return false;
}


void printc(const Complex c[2])
{
	printf("(%+.3f %+.3fi) --- (%+.3f %+.3fi)\n",
		real(c[0]), imag(c[0]), real(c[1]), imag(c[1]));
}



class chord
{
	public:		// pick 6 magic numbers for a chord
		chord();
		void next(double ch[6]);

	private:
		int n;	// time
		int jj[6];

};


chord::chord()
{
	for(int i=0; i<6; i++)
		jj[i] = i;
}


void chord::next(double ch[6])
{
	for(int i=0, k; i<6; i++)
		{
		k = jj[i];
		ch[i] = constants::magic_numbers[k];
		//printf("%d -- jj %d -> %.5f\n", i, k, ch[i]);
		if( (jj[i] += 1+i) >= 31 )
			jj[i] -= 31;
		}
}


class siren
{
	public:
		siren(int sr=48000);
		~siren() {osc1=0; osc2=0; delete osc2; delete osc2; wtb=0; delete wtb;}
		double slide();
		void tune();

	private:
		int sr;
		double A1, A2;
		double x, y;
		double f1, f2;
		oscil *osc1, *osc2;
		double *wtb;
		double phi1, phi2;
		double W;
		double LF[6];
		int n;
		double mod;
		long int timer;
};


siren::siren(int fs)
{
	sr = fs;
	W = 2*M_PI/sr;
	timer = 0;

	for(int i=0; i<6; i++)
		LF[i] = 1000 * constants::magic_numbers[18+i];

	mod = 1.0;
	A1 = A2 = 0.0;
	const int L = 1025;
	wtb = new double[L];
	waveform::make_sine(wtb, L);
	osc1 = new oscil(sr, wtb, L);
	osc2 = new oscil(sr, wtb, L);
	phi1 = phi2 = 0;
	x = y = 0;
}


void siren::tune()
{
	if(++n >= 6)
		n = 0;

	if(timer > 3*sr)
		{
		mod = 2.5;
		timer = 0;
		}
}

double siren::slide()
{
	float frq = 23.5 * (0.5 + mod) + LF[n];
	y = osc2->stable(frq * (1 + y*y) + mod*x);
	x = osc1->stable(LF[n] * (1 + mod*y + fabs(x)));

	mod *= .99985;
	A1 = .25*tanh(mod);
	++timer;
	return A1*x + .05*y;
}


void play()
{
	const unsigned char ch = 4;
	const int sr = 48000;
	SNDFILE *snafu;
	SF_INFO flake;
	flake.samplerate = sr;
	flake.channels = ch;
	flake.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	snafu = sf_open("quasiperiodic_table.wav", SFM_WRITE, &flake);

	const int LL = 1 << 8;

	//double buf[ch*LL];
	double *buf;
    buf = new double[ch*LL];

	const int TT = 175 * sr / LL;	// 275

	quasiperiod Haken;
	siren Xenon(sr);
	Complex z[2];
	z[0] = Complex(1.0/sqrt(7), 0), z[1] = Complex(1.0/sqrt(5), 0);

	double g = 1/4.;

	slope salope;
	bool PR;

	double frq[6] = {sqrt(3), 1/sqrt(2), sqrt(5), constants::phi, constants::_d, constants::_R};
	chord barbershop;
	long k=0, timer=0;

	double x;
	int clip1=0, clip2=0;

	for(int j=0; j<TT; j++)
		{
		for(int i=0; i<ch*LL; i += ch, k++)
			{
			Haken.rotate(z);
			x = Xenon.slide();

			buf[i  ] = g*real(z[0]);
			buf[i+1] = g*imag(z[0]);
			buf[i+2] = g*real(z[1]) + .05*x;
			buf[i+3] = g*imag(z[1]) + .05*x;
			if(fabs(buf[i]) > 1 || fabs(buf[i+1]) > 1)
				++clip1;
			if(fabs(buf[i+2]) > 1 || fabs(buf[i+3]) > 1)
				++clip2;

			PR = salope.rising(real(z[1]));
			if((PR && k%sr < 20) || ++timer > 7*sr)
				{
				if(Haken.tune(frq))
					{
					barbershop.next(frq);
					timer = 0;
					}
				Xenon.tune();
				}
			}
		sf_write_double(snafu, buf, ch*LL);
		}

	sf_close(snafu);

	printf("clip: %d + %d \n", clip1, clip2);
    buf = 0;
    delete buf;
}





int main()
{
	play();
	return 0;
}


