/*
	hello world

	(c) Risto Holopainen 2021
	Output a coded "Hello World!" message as soundfile
	compile with gcc hello.c -lsndfile -o hello

	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "hello.h"
#include "parabolic.h"
#include <stdio.h>
#include <stdlib.h>
#include <sndfile.h>


char alphabet[AN];	// for global communication


void transform(const char *msg, int sz, int Tr, char *out)
{
	/*
	msg : int 0..AN-1
	rotate char by Tr mod AN
	*/
	int i=0;
	int cc;
	while(i < sz)
		{
		cc = (msg[i] + Tr) % AN;
		out[i] = alphabet[cc];
		// printf("t%2d cc %2d --> %c \n", i, cc, alphabet[cc]);
		++i;
		}
	out[ML] = '\0';
}


void to_key(const char *msg, int sz, char *out)
{
	/* char --> 0..35 */
	int i=0, j=0;
	while(i < sz)
		{
		if(msg[i] >= 'a' && msg[i] <= 'z')
			out[i] = msg[i] - 'a' + 1;
		else if(msg[i] == '_')
			out[i] = 0;

		for(j=27; j<AN; j++)
			if(msg[i] == alphabet[j])
				{
				out[i] = j;
				break;
				}
		++i;
		}
}


double to_param(const double *scale, const char *text, int sz, int rot, parameters *PL)
{
	// return end-time
	/* text: 0..35; rot0--12 */

	double f, a, a1 = 0.5, b, t;
	char c;
	int i, j;
	const double MA = 1.0 / (AN*AN);
	const double gamma = 0.93;
	const double eps = 1.0 - gamma;
	double start = PL[0].st;
	char ch = PL[0].chn;
	parameters pm;
	pm.st = start;

	for(i=0; i<sz; i++)
		{
		j = (i + rot) % ML;
		c = text[j] + 1;
		b = 1.0/c + MA*c*c;
		a = b + a1;
		while(a > 1.0)
			a = gamma * a - eps;

		if (c == 27)
			a = eps;
		a1 = a;
		t = 0.5 * (1 + 1/(2*b+1));
		f = scale[c-1];
		pm.frq = f;
		pm.amp = a;
		pm.dur = t;
		pm.chn = ch;
		PL[i] = pm;
		pm.st += t;
		}

	return PL[sz-1].st + PL[sz-1].dur;
}


void synt(const parameters *u, const parameters *v, int sz, osc *Os1, osc *Os2, SNDFILE *snd)
{

	long t, Tu, Tv, endtime;
	double e1, e2;
	double x, y;
	double frq_u = u[0].frq, frq_v = v[0].frq, amp_u = u[0].amp, amp_v = v[0].amp;
	unsigned int M = sz-1;
	int m=0, n=0;
	unsigned int i=0;
	t = u[0].st;
	const unsigned int bufsz = 1 << 8; //256
	//double buf[bufsz];
	double buf[1 << 8];// MSV debug

	static double au1=0, av1=0;
	const double decay = 0.975, eps = 0.005;

	Tu = u[0].st + u[0].dur;
	Tv = v[0].st + v[0].dur;
	e1 = u[M].st + u[M].dur;
	e2 = v[M].st + v[M].dur;

	endtime = (e1 < e2) ? SR*e1 : SR*e2;
	//printf("synt: endtime = %ld (%.3f sec)\n", endtime, (float) endtime / SR);
	//printf("e1: %.1f s \te2: %.1f s\n", e1, e2);

	while(t++ < endtime)
		{
		amp_u = .25*amp_u + .75*au1;
		amp_v = .25*amp_v + .75*av1;

		au1 = amp_u;
		av1 = amp_v;

		x = amp_u*parabolic_tri(frq_u, Os1);
		y = amp_v*parabolic_tri(frq_v, Os2);

		buf[i  ] = x;
		buf[i+1] = y;
		i += 2;
		if(i >= bufsz)
			{
			sf_write_double(snd, buf, bufsz);
			i = 0;
			amp_u = amp_u*decay + eps;
			amp_v = amp_v*decay + eps;
			}

		if(t >= Tu)
			{
			Tu += SR * u[m].dur;
			++m;
			frq_u = u[m].frq;
			amp_u = u[m].amp;
			}

		if(t >= Tv)
			{
			Tv += SR * v[m].dur;
			++n;
			frq_v = v[m].frq;
			amp_v = v[m].amp;
			}

		if(m >= sz || n >= sz)
			break;
		}
	// flush rest of buffer
	sf_write_double(snd, buf, i);

	//printf("exit-synt: t %5ld, Tu %3ld, Tv %3ld, m: %2d, n: %2d\n",
	//	t, Tu, Tv, m, n);

}


int main()
{
	make_alphabet(alphabet);
	signed int n=0;
	char out[ML+1];
	char trs[ML+1];
	char KEY[ML+1];
	int offset=0;
	double scale[AN];
	make_scale(scale, AN);
	parameters plx[ML], ply[ML];
	plx[0].chn = 0;
	ply[0].chn = 1;
	double endx, endy;

	osc Osc1, Osc2;
	init_osc(&Osc1, SR);
	init_osc(&Osc2, SR);

	SNDFILE *ouf;
	SF_INFO head;
	head.samplerate = SR;
	head.channels = 2;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	ouf = sf_open("hello_world.wav", SFM_WRITE, &head);

	to_key(msg, ML, trs);
	n = -1;
	while(++n < ML)
		{
		offset = trs[n]+7;
		transform(trs, ML, offset, out);
		to_key(out, ML, KEY);
		printf("; transf. %2d \t %s \t o: %2d\n", n, out, offset);
		endx = to_param(scale, KEY, ML, n, plx);
		endy = to_param(scale, KEY, ML, 0, ply);

		plx[0].st = endx;
		ply[0].st = endy;

		synt(plx, ply, ML, &Osc1, &Osc2, ouf);
		}

	sf_close(ouf);
	return 0;
}


