

/*
	origami
	Copyright Risto Holopainen 2021

	forward-backward sinusoid trajectory with wrinkles
	For research, educational purposes, and for the curious.
	Running this program produces a three channel soundfile which is
	an algorithmic composition.


	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#ifndef M_PI
#define M_PI 3.14159265358979323846 // instead of #define _USE_MATH_DEFINES
#endif


#include <cstdio>
#include <cmath>
#include <vector>
#include <sndfile.h>
#include "origami.h"

using std::vector;

enum direction {Forward = 1, Backward = -1};

struct timeconst
{
	unsigned int sr;
	double duration_seconds;
	unsigned long totdur;
	timeconst()
		{sr=48000; duration_seconds=0; totdur=0;}
};


struct Parameters
{
	double slope;		// glissando
	double frq;
	double glr;			// gliss-rate
	direction dir;		// +/- 1
	unsigned char chn;	// enum?
	double amp;			// 0-1
	double dur;			// sec.
	double time;		// current pos. sec.
	timeconst tm;		// samplerate & duration

	Parameters(double Sl, double Fr, double Gr, direction Dr,
		unsigned char C, double A, double D, double T, timeconst Tc)
		: slope(Sl), frq(Fr), glr(Gr), dir(Dr),
		chn(C), amp(A), dur(D), time(T), tm(Tc) {};

	Parameters()
		{slope=0; frq=6000; glr=0; dir=direction::Forward; chn=0; amp=0.0; dur=0.0; time=0.0;};
};


struct tape
{
	timeconst tm;
	vector<double> L;
	vector<double> C;
	vector<double> R;
};


Parameters map(const Parameters &pp)
{
	Parameters qq = pp;

	double s = pp.slope;
	double a = pp.amp;
	double d = pp.dur;
	int dir = pp.dir;
	unsigned char cc = pp.chn;
	double sqf = sqrt(fabs(pp.frq));
	double s1, a1, d1, fr;
	s1 = sin(M_PI*s) + origamix::alpha*a - 1.0/(1 - 1/sqf + origamix::beta*cc);
	a1 = exp(-d) + origamix::gymel*a;
	d1 = origamix::contract*d + origamix::qoph*(s*s + a*a) + origamix::mumu / sqf;
	fr = sqf*s + pp.frq;
	qq.slope = s1;
	qq.dir = (a < 0.25) ? Forward : Backward;

	if(pp.dir == Backward)
		cc = (cc+1) % 3;

	qq.chn = cc;
	qq.amp = a1;
	qq.dur = d1;
	qq.frq = fr;
	qq.time = pp.time + dir*d1;
	while(qq.time < 0) qq.time += pp.tm.duration_seconds;
	while(qq.time >= pp.tm.duration_seconds) qq.time -= pp.tm.duration_seconds;

	return qq;
}


void synthesis(tape &TP, const Parameters &p)
{
	const int sr = TP.tm.sr;
	const int incr = (int)p.dir;
	long int tp = sr * p.time;	// anchor time point
	const int duration = sr*p.dur;
	unsigned long tau;
	const double W = 2.0*M_PI/sr;
	double frq = p.frq;
	const double sqf = sqrt(fabs(p.frq));
	double gl = sqf * p.slope / duration;
	double x;
	static double theta = 0;
	const double decay = 1.0 - 3.25E-05;
	double amp = p.amp;
	const unsigned long tot = p.tm.totdur;

	for(int j=0, k=0; j<duration; j++, k += incr)	// bidirectional iteration
		{
		// synthesize
		tau = (tp + k + tot) % tot;
		frq += gl;
		theta += W*frq;
		x = sin(theta);		// might as well use wavetable lookup for efficiency

		if(p.chn == 1)
			TP.C[tau] += amp * x;
		else if(p.chn == 0)
			TP.L[tau] += amp * x;
		else if(p.chn == 2)
			TP.R[tau] += amp * x;

		amp *= decay;
		}

	// printf("synth frq %.1f, \tdf %.2f, \tgl %.5f \n", frq, sqf, gl);
}



void init_tape(tape &tp, const timeconst &tc)
{
	tp.tm = tc;
	const int N = tc.totdur;
	tp.L.reserve(N);
	tp.C.reserve(N);
	tp.R.reserve(N);
	// record high beep on tape, lopass to remove
	double eps = 0.0025, u;

	for(int i=0, j=1; i<N; i++)
		{
		j = ((i&1) << 1) - 1;
		u = eps*j;
		tp.L[i] =  u;
		tp.C[i] = -u;
		tp.R[i] =  u;
		}
}


void normalize(tape &tp)
{
	const long N = tp.tm.totdur;
	double maxamp = 0.0, x, y, z;

	for(long i=0; i<N; i++)
		{
		x = tp.L[i];
		y = tp.C[i];
		z = tp.R[i];
		if( fabs(x) > fabs(maxamp) )
			maxamp = x;
		else if(fabs(y) > fabs(maxamp))
			maxamp = y;
		else if(fabs(z) > fabs(maxamp))
			maxamp = z;
		}

	double norm = 1.0 / maxamp;
	printf("max_amp C: %.3f, norm = %.4f \n",
		maxamp, norm);

	for(long i=0; i<N; i++)
		{
		tp.L[i] *= norm;
		tp.C[i] *= norm;
		tp.R[i] *= norm;
		}
}


void to_sfile(const tape &tp)
{
	SNDFILE *sfo;
	SF_INFO head;

	head.samplerate = tp.tm.sr;
	head.channels = 3;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	sfo = sf_open("Origamixd3.wav", SFM_WRITE, &head);

	const int L = 1 << 10;
	const int T = tp.tm.totdur / L;
	double buf[3*L];

	for(long t=0, k=0; t<T; t++)
		{
		for(int i=0; i<3*L; i += 3, k++)
			{
			buf[ i ] = .75*tp.L[k];
			buf[i+1] = .75*tp.R[k];
			buf[i+2] = .75*tp.C[k];
			}
		sf_write_double(sfo, buf, 3*L);
		}

	sf_close(sfo);
}


void set_timeconstants(timeconst &tt, double dur_sec=30)
{
	tt.sr = 48000;
	tt.duration_seconds = dur_sec;
	tt.totdur = tt.sr * tt.duration_seconds;
}

int main()
{
	timeconst TM;
	set_timeconstants(TM, origamix::length_of_composition);
	tape snd;
	init_tape(snd, TM);
	Parameters pp(0, TM.sr/12, 0.0, Forward, 0, 0.5, 1.0, 0, TM);
	printf("TM.totdur = %ld\n", TM.totdur);

	const int N_iterations = origamix::all_those_iterations;	// of the map
	int n = 0;
	while(++n < N_iterations)
		{
		pp = map(pp);
		printf("%2d \t t %.2f \t dir %+0d \t slope %+.2f \tamp %.3f \tdur %.3f \t %.1f Hz \tch %d\n",
			n, (float)pp.time, (int)pp.dir, pp.slope, pp.amp, pp.dur, pp.frq, pp.chn);

		synthesis(snd, pp);
		if(fabs(pp.frq) > snd.tm.sr/3)
			break;
		}

	normalize(snd);
	to_sfile(snd);
	return 0;
}





