/*
	timbre.h
	Copyright Risto Holopainen 2021
	
	help functions to generate the Wastebasket of Timbre
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>. 

*/

#include "d_tuning.h"
#include "../util/sequencer.h"
#include "../util/oscil.h"
#include "../util/biquad2.h"
#include "../util/waveforms.h"

typedef deque<double> pitchline;

void make_seq_G(pitchline &frqs, int N, int m)
{
	d_tuning DT;
	double F;
	int pc, oct;
	frqs.clear();
		
	for(int i=m, k=0; i<N+m; i++)
		{
		++k %= 5;
		pc = k + i;
		oct = (i%4) - 2;
		F = DT.frq(pc, oct, d_tuning::G);
		frqs.push_back(F);
		}
}


void make_seq_Ciss(pitchline &frqs, int N, int m)
{
	d_tuning DT;
	double F;
	int pc, oct[5] = {2,0,1,-1,0};
	frqs.clear();
	
	for(int i=m, k=0; i<N+m; i++)
		{
		++k %= 5;
		pc = i + m;
		F = DT.frq(pc, oct[k], d_tuning::Ciss);
		frqs.push_back(F);
		}
}


void portamento(const double frq[2], double F[2], float w)
{
	static double a=0.01, b=1-a;
	static one_pole<double> porta_1(a, b), porta_2(a, b);
	if(w > 0.4 && w < 0.74)
		{
		double q = w-0.4;
		a = 0.02 + 2*q;
		b = 1-a;
		porta_1.set_coef(a,b);
		}
	else if(w > 0.75 && a > 0.003)
		{
		a *= 0.995;
		b = 1-a;
		porta_2.set_coef(a,b);
		}

	F[0] = porta_1.filter(frq[0]);
	F[1] = porta_1.filter(frq[1]);
}



using namespace waveform;
class vibrato
{
	public:
		vibrato(int kr);
		~vibrato();
		double vib(float w);
		
	private:
		int kr;
		int L;
		oscil *lfo;
		double *wt;
};

vibrato::vibrato(int fs)
{
	kr = fs;
	L = 513;
	wt = new double[L];
	make_triangle(wt, L, 7);
	lfo = new oscil(kr, wt, L);
}

vibrato::~vibrato()
{
	wt = 0;
	lfo = 0;
	delete wt;
	delete lfo;
}


double vibrato::vib(float w)
{
	double u = 24*w;
	double m = u - floor(u);
	double a = 4*w*(1-w);
	double x = a * lfo->stable(2 + 5*m);
	return x;
}




class orchestra
{
	public:
		orchestra(int fs, 
				const double *w1, const double *w2, 
				const double *w3, const double *w4, int sz);
		~orchestra();
		void toot(const double frqs[4], double x[4]);

	
	private:
		int sr;
		int L;
		oscil *osc1, *osc2, *osc3, *osc4;
		double *wt1, *wt2, *wt3, *wt4;
};


orchestra::orchestra(int fs, 
				const double *w1, const double *w2, 
				const double *w3, const double *w4, int sz)
{
	/*
	def wavetables externally!
	*/
	sr = fs;
	L = sz;
	wt1 = new double[L];
	wt2 = new double[L];
	wt3 = new double[L];
	wt4 = new double[L];
	// make internal copy of waves
	for(int i=0; i<sz; i++)
		{
		wt1[i] = w1[i];
		wt2[i] = w2[i];
		wt3[i] = w3[i];
		wt4[i] = w4[i];
		}
	
	osc1 = new oscil(sr, wt1, L);
	osc2 = new oscil(sr, wt2, L);
	osc3 = new oscil(sr, wt3, L);
	osc4 = new oscil(sr, wt4, L);
}


orchestra::~orchestra()
{
	wt1 = wt2 = wt3 = wt4 = 0;
	delete wt1; delete wt2;
	delete wt3; delete wt4;
	osc1 = osc2 = osc3 = osc4 = 0;
	delete osc1; delete osc2; delete osc3; delete osc4;
}



void orchestra::toot(const double frq[4], double x[4])
{
	x[0] = osc1->stable(frq[0]);
	x[1] = osc2->stable(frq[1]);
	x[2] = osc3->stable(frq[2]);
	x[3] = osc4->stable(frq[3]);
}



class quark	// quad-sequencer : 4 pitches
{
	public:
		quark(int p=0);
		~quark();
		void step(double frq[4]);	// and permutate
		
	private:
		sequencer<double> *a, *b, *c, *d;
		int k;
		double oct;	// ratio
};

quark::quark(int p)
{
	k = 0;
	pitchline s1, s2, s3, s4;

	make_seq_G   (s1,  7, 1+p);
	make_seq_Ciss(s2,  8, 2+p);
	make_seq_G   (s3,  9, 3+p);
	make_seq_Ciss(s4, 11, 5+p);

	a = new sequencer<double> (s1);
	b = new sequencer<double> (s2);
	c = new sequencer<double> (s3);
	d = new sequencer<double> (s4);
	oct = 1.0;
}

quark::~quark()
{
	a = b = c = d = 0;
	delete a; delete b; delete c; delete d;
}

void quark::step(double F[4])
{
	F[0] = a->tick();
	F[1] = b->tick();
	F[2] = c->tick();
	F[3] = d->tick();

	++k;
	F[k%4] *= oct;
	
	oct = 1.0;
	if(k%13 == 0)
		d->swap(1, k%11);
	if(k%11 == 1)
		{
		c->swap(0, k%8 + 1);
		oct = 1/4.;
		}
	if(k%17 == 2)
		{
		b->reverse();
		oct = 1/2.;
		}
	if(k%19 == 3)
		{
		a->stride(1 + k%6);
		oct = 2;
		}
}



class quadfilter
{
	public:
		quadfilter(int sr);
		~quadfilter();
		void filter(const double x[4], double st[2]);

	private:
		int sr;
		double y[4];
		bandpass *bp1, *bp2, *bp3, *bp4;
		float cfrq[4];
};

quadfilter::quadfilter(int fs)
{
	sr = fs;
	bp1 = new bandpass(sr, 450, 0.7);
	bp2 = new bandpass(sr, 540, 0.7);
	bp3 = new bandpass(sr, 650, 0.7);	
	bp4 = new bandpass(sr, 780, 0.7);

	for(int i=0; i<4; i++)
		{
		cfrq[i] = 450 + 85*i + 3*i*(i+1);
		y[i] = 0.0;
		}
}


quadfilter::~quadfilter()
{
	bp1 = bp2 = bp3 = bp4 = 0;
	delete bp1; delete bp2; delete bp3; delete bp4;
}


void quadfilter::filter(const double x[4], double st[2])
{
	y[0] = bp1->filter(x[0]);
	y[1] = bp2->filter(x[1]);	
	y[2] = bp3->filter(x[2]);
	y[3] = bp4->filter(x[3]);

	st[0] = .5*(y[0] + y[2] + .5*(x[0] + x[1]));
	st[1] = .5*(y[1] + y[3] + .5*(x[2] + x[3]));
}
